!+
!PURPOSE:
!    Create a mock CMB data set file C_l 
!AUTHOR:
!    L. Perotto
!OUTPUTS:
!    The datafile is organized in 10 columns as follows:
!    ell | ClTT | ClTE | ClEE | Cldd | ClTd | NlT | NlP | Nld | fsky
!    spectra are given without any l(l+1)/2pi factor
!    ClTT,ClTE,ClEE are in muK^2
!    ClTd (CMB temp-deflection angle) in muK
!    Cldd (deflection angle) no unit   
!
!DEPENDANCIES:
!    utils.F90 file in cosmomc package (AMLutils module)  
!    inifile.f90 file in cosmomc (IniFile module)
!    FUTURCMB_random.f90 written ky Y.Y.Y. Wong
!    FUTURCMB_lensnoise.f90 written by J. Lesgourgues
!
!VERSION:
!    nov 2009: fully revised to match recent cosmomc update
!    sep 2006: new version: a spurious Ed correlation is killed. 
!              a multivariate random number generator 
!              written by Y.Y.Y. Wong is used
!    jun 2006: input param. file + cosmethics by LP
!    feb 2006: written by LP  
!-


program FuturCMB
  use AMLutils
  use IniFile
  use lensnoise
  use random_numbers
  
  implicit none
  
  !  DECLARATIONS
  !_____________________________________________________________________
  
  CHARACTER(LEN=*), PARAMETER :: code = 'FuturCMB'
  real, parameter :: Tcmb = 2.726e6  ! micro K
  character(LEN=100) ParamFile
  
  character(LEN=100) :: setname
  character(LEN=500) :: datarep
  
  !! CMB part
  logical :: CMBonly
  integer, parameter :: jjmaxTmax=100 ! calc_Nldd
  character(LEN=500) :: cmbcl_infile
  character(LEN=500) :: cmbcllens_infile
  integer lmin, lmax, num_cls, nchan, nc
  
  real, dimension(:), allocatable :: fwhm_arcmin, sigma_T, sigma_P
  real xlc
  real, dimension(:), allocatable :: fwhm_rad, sigma2
  real T, E, TE, B, kk, Tk, ndd, d, Td
  character(LEN=500) InLine
  double precision ell, scal
  
  REAL, dimension(:,:), allocatable :: Clfid
  REAL, dimension(:,:), allocatable :: InCls
  REAL, dimension(:,:), allocatable :: InlensCls
  REAL, dimension(:,:), allocatable :: NoiseCl
  double precision, dimension(:), allocatable :: nldd
  
  integer :: ldum,i,j,l,ll
  real :: fsky
  
  integer :: seed
  
  logical :: DoMock
  integer chatty
  integer bonalloc 
  logical bad
  
  ! CMB covariance matrix
  double precision, dimension(:,:,:), allocatable :: Bigcov
  double precision, dimension(:,:,:), allocatable :: Bigcov_mock
  integer :: n
  
  character(LEN=500) :: outname, setoutname
  
    
  
  !__________________________________________________________________
  
  
  
  
  ! Is the parameter input file here?
  !________________________________________________________
  if (iarg() < 1) then 
     print *, code// "> The parameters input file is missing"
     stop
  else
     call getarg(1,ParamFile)
  endif
  
  
  ! Reading of the parameter input file
  !________________________________________________________
  call Ini_Open(ParamFile, 1, bad, .false.)
  if (bad) then 
     print*, code// '> Error opening parameter file'
     stop
  endif
  Ini_fail_on_not_found = .false.
  
  setname = Ini_Read_String('setname')
  datarep = Ini_Read_String('datarep')
  cmbcl_infile = Ini_Read_String('cmbcl_infile')
  cmbcllens_infile = Ini_Read_String('cmbcllens_infile')
  DoMock = Ini_Read_Logical('DoMock')
  chatty = Ini_Read_Int('chatty',0)
  lmin = Ini_Read_Int('lmin')
  lmax = Ini_Read_Int('lmax')
  fsky = Ini_Read_Real('fsky')
  num_cls = Ini_Read_Int('num_cls')
  nchan = Ini_Read_Int('nchan')
  !seed = Ini_Read_Negint('seed',-90)
  seed = Ini_Read_Int('seed',90)

  !
  ! test of chosen number of spectra
  !________________________________________________________
  if (num_cls < 3) then 
     num_cls = 3
     if (chatty > 0) then
        print*, code// "> temperature AND polarisation spectra will appear in the output file"
     endif
  endif
  if (num_cls == 4) then
     print*, code// "> number of spectra should be either 3 or 5"
     print*, code// "> please modify the input param. file."
     stop
  endif
  if (num_cls > 5) then 
     num_cls = 5
     if (chatty > 0) then
        print*, code// "> The output file will contain TT, EE, TE, dd and Td spectra"
        print*, code// "> Primordial B-mode is assumed to be zero..."
     endif
  endif
  
!!_________________________________________
!!
!! CMB part
!!__________________________________________

  ! experimental model
  !________________________________
  if (nchan > 0) then
     allocate(fwhm_arcmin(1:nchan), &
          &        sigma_T(1:nchan), &
          &        sigma_P(1:nchan), &
          &        fwhm_rad(1:nchan),&
          &        sigma2(1:nchan),stat = bonalloc)
     InLine = Ini_Read_String('fwhm_arcmin',.true.)
     read(InLine, *) fwhm_arcmin(1:nchan)
     InLine = Ini_Read_String('sigma_T', .true.) 
     read(InLine, *) sigma_T(1:nchan)
     InLine = Ini_Read_String('sigma_P', .true.) 
     read(InLine, *) sigma_P(1:nchan)
     xlc= sqrt(8.*log(2.))
     fwhm_rad(:) = fwhm_arcmin(:)/60.*pi/180.
     sigma2(:) = (fwhm_rad(:)/xlc)**2
  endif
  
  !  The fiducial CMB spectra (C_l^{theo} + N_l)
  !______________________________________________________
  allocate(Clfid(1:num_cls,lmin:lmax), &
       &   InCls(1:num_cls,1:lmax), &
       &   NoiseCl(2,1:lmax), stat = bonalloc)
  if (bonalloc > 0) then
     print*,code//'> pb allocation memoire', bonalloc
     stop
  endif
  if (num_cls == 5) then
     allocate(InlensCls(4,1:lmax), stat = bonalloc)
     if (bonalloc > 0) then
        print*,code//'> pb allocation memoire', bonalloc
        stop
     endif
  endif
  
  Clfid = 0.
  InlensCls = 0.
  InCls = 0.
  NoiseCl = 0.                                         
  !    The CAMB files are read   
    
  ! No tensor mode

  ! N_l is non-zero
  if (nchan < 1.) then
     print*,"Experiment has a channel at least..."
     stop
  endif 

  if (num_cls == 3) then
     open(unit=26,file=cmbcl_infile,form='formatted',status='old')
     do ll = lmin, lmax
        read (26,*) l, T, E, TE
        if (l/=ll) then
           print *,code//'> cl must start at l = ', lmin,' and end at l = ',lmax 
           stop 
        endif
        ell = real(ll,KIND(1.d0))
        scal = twopi/(ell*(ell+1))
        
        do nc = 1,nchan
           NoiseCl(1,ll) = NoiseCl(1,ll) +     &
                1./((fwhm_rad(nc)*sigma_T(nc))**2) &
                *exp(real(-ll*(ll+1),KIND(1.d0))*sigma2(nc))
           NoiseCl(2,ll) = NoiseCl(2,ll) +     &
                1./((fwhm_rad(nc)*sigma_P(nc))**2) &
                *exp(real(-ll*(ll+1),KIND(1.d0))*sigma2(nc))
        enddo
        
        NoiseCl(1,ll) = 1./NoiseCl(1,ll)
        NoiseCl(2,ll) = 1./NoiseCl(2,ll)
        
        InCls(1,ll) = T*scal  ! micro K^2
        InCls(2,ll) = E*scal  ! micro K^2
        InCls(3,ll) = TE*scal ! micro K^2 
        
        Clfid(1,ll) = T*scal + NoiseCl(1,ll) ! micro K^2
        Clfid(2,ll) = E*scal + NoiseCl(2,ll) ! micro K^2
        Clfid(3,ll) = TE*scal ! micro K^2
                
     enddo
     close(26)
  endif

  if (num_cls == 5) then
     open(unit=26,file=cmbcl_infile,form='formatted',status='old')
     do ll = lmin, lmax
        read (26,*) l, T, E, TE, kk, Tk
        if (l/=ll) then
           print *,code//'> cl must start at l = ', lmin,' and end at l = ',lmax 
           stop 
        endif
        ell = real(ll,KIND(1.d0))
        scal = twopi/(ell*(ell+1))
        
        do nc = 1,nchan
           NoiseCl(1,ll) = NoiseCl(1,ll) +     &
                1./((fwhm_rad(nc)*sigma_T(nc))**2) &
                *exp(real(-ll*(ll+1),KIND(1.d0))*sigma2(nc))
           NoiseCl(2,ll) = NoiseCl(2,ll) +     &
                1./((fwhm_rad(nc)*sigma_P(nc))**2) &
                *exp(real(-ll*(ll+1),KIND(1.d0))*sigma2(nc))
        enddo
        
        NoiseCl(1,ll) = 1./NoiseCl(1,ll)
        NoiseCl(2,ll) = 1./NoiseCl(2,ll)
        
        InCls(1,ll) = T*scal  ! micro K^2
        InCls(2,ll) = E*scal  ! micro K^2
        InCls(3,ll) = TE*scal ! micro K^2 
        InCls(4,ll) = kk*(ell+1)/ell**3/(Tcmb)**2   ! no unit
        InCls(5,ll) = Tk*sqrt(ell*(ell+1))/ell**3/(Tcmb) ! micro K
        
        Clfid(1,ll) = T*scal + NoiseCl(1,ll) ! micro K^2
        Clfid(2,ll) = E*scal + NoiseCl(2,ll) ! micro K^2
        Clfid(3,ll) = TE*scal ! micro K^2
        ! Cldd
        Clfid(4,ll) = kk*(ell+1.)/ell**3/Tcmb**2   ! no-unit 
        Clfid(5,ll) = Tk* &
             & sqrt((ell+1.)*ell)/ell**3/Tcmb ! micro K
        
     enddo
     close(26)
  endif
  
  ! save memory
  if (nchan > 0) deallocate(fwhm_arcmin, &
       &        sigma_T, &
       &        sigma_P, &
       &        fwhm_rad,&
       &        sigma2)
  
  
  !     N_l^{dd} calculation 
  !    (following the algorithm given in Okamoto & Hu)
  !______________________________________________________________
  if (num_cls == 5) then

     if (chatty > 0)  print * , code//'> N_l^dd calculation'
     open(unit=1,file=cmbcllens_infile,form='formatted',status='old') 
     do ll = lmin, lmax
        read (1,*) l, T, E, B, TE
        if (l/=ll) then
           print *,code//'> cl must start at l = ', lmin," and end at l = ",lmax
           stop 
        endif
        ell = real(ll,KIND(1.d0))
        scal=twopi/(ell*(ell+1)) 
        InlensCls(1,ll) = T*scal
        InlensCls(2,ll) = E*scal
        InlensCls(3,ll) = TE*scal
        InlensCls(4,ll) = B*scal
     enddo
     close(1)
     
     allocate(nldd(1:lmax), stat = bonalloc)
     if (bonalloc > 0) then
        print*,code//'> pb of memory allocation ', bonalloc
        stop
     endif
     do ll = 1, lmax
        nldd(ll)=0.
     enddo
     
     if (lmin > 1) then
        InCls(1:num_cls,1:lmin-1) = 0.
        InlensCls(1:4,1:lmin-1) = 0.
        NoiseCl(1:2,1:lmin-1) = 0.
     endif
     
     !test
     
     call calc_Nldd(InCls, InlensCls, NoiseCl, nldd, jjmaxTmax, lmax)
     
     do ll = lmin, lmax
        ell = real(ll)
        Clfid(4,ll) = Clfid(4,ll) + nldd(ll) ! deflection angle
     enddo
     
     !save memory
     deallocate(InlensCls)
     
  endif
  
  !save memory
  deallocate(InCls)

  
  !    ---- check input power spectra consistency  ----
  
  do i=1,num_cls
     do l=lmin,lmax
        if ((i < 3) .or. (i == 4)) then ! TT, EE, dd
           if (Clfid(i,l) < 0.0) then
              print*,code,'> Negative input power spectrum at l =', l,', index = ',i
              print*,code,'> ',Clfid(i,l)
              stop
           endif
        endif
     enddo
  enddo
  do l=lmin,lmax
     T=Clfid(1,l)
     E=Clfid(2,l)
     Td=Clfid(5,l)
     TE=Clfid(3,l)
     d=Clfid(4,l)
     if (abs(TE) > sqrt(T)*sqrt(E) ) then
        print *,code,'> Inconsistent T, E and ExT power spectrum terms at l = ',l
        print*,code,'> ',T,E,TE
     else ! Td
        if (abs(Td) > sqrt(d)*sqrt(T - (TE**2/E))) then
           print *,code,'> Inconsistent T, d and Txd power spectrum terms at l = ',l
           print*,code,'> ',Td,T,d,TE,E
        endif
     endif
  enddo
  
!!_________________________________________
!!
!! All together in the covariance matrix !!
!!_________________________________________
  
  if (num_cls == 3) n=2
  if (num_cls == 5) n=3
  if (chatty > 0) print*,code//"> Cov. matrix dim. = ",n
  allocate(Bigcov(lmin:lmax,n,n), &
       &  stat = bonalloc)
  if (bonalloc > 0) then
     print*,code//'> pb allocation memoire', bonalloc
     stop
  endif

  Bigcov(lmin:lmax,1:n,1:n) = 0.
  
  if (num_cls == 5 ) then
     do l = lmin, lmax  
        Bigcov(l,1,3) = Clfid(5,l) !Td
        Bigcov(l,3,3) = Clfid(4,l) !dd
     end do
  endif
  do l = lmin, lmax   
     Bigcov(l,1,1:2) = Clfid((/1,3/),l)
     Bigcov(l,2,2) = Clfid(2,l)
     !symetrisation
     do i=1, n
        do j=i+1, n
           Bigcov(l,j,i) = Bigcov(l,i,j)
        end do
     end do
     
  end do
  
  !save memory
  deallocate(Clfid)
    
  
!!!!!!!!!!!!!  Generate random numbers  !!!!!!!!!!!!!!!!!!!
  
  if (DoMock == .true.) then
     allocate(Bigcov_mock(lmin:lmax,n,n), &
          &  stat = bonalloc)
     if (bonalloc > 0) then
        print*,code//'> pb allocation memoire', bonalloc
        stop
     endif
     
     !seed must be a negative integer.
     seed = seed * (-1)
     
     if (chatty > 0) print*,code//"> Generate random numbers..."
     if (chatty > 0) print*,code//"> seed = ",seed
     

      
     call get_random_cls(n,lmin,lmax,seed,Bigcov,Bigcov_mock)
     
  else  
     if (chatty == 1) print *, code//'> filling dataset with fiducial spectra' 
  endif
  
  !
  ! Output files
  !_____________________________________________________
  outname = trim(datarep)//trim(setname)//".dat"
  if (chatty > 0) print *, code//'> writing the output file: '//trim(outname)

  call CreateTxtFile(outname,34)
  if (num_cls == 5) then
     if (DoMock == .false.) then 
        do l=lmin, lmax     
           write (34,'(1I5,9E15.5)') l, Bigcov(l,1,1:2), Bigcov(l,2,2), &
                &  Bigcov(l,3,3), Bigcov(l,1,3), &
                &  NoiseCl(1:2,l), Nldd(l), fsky
        end do
     else
        do l = lmin, lmax     
           write (34,'(1I5,9E15.5)') l, Bigcov_mock(l,1,1:2), Bigcov_mock(l,2,2), &
                &  Bigcov_mock(l,3,3), Bigcov_mock(l,1,3), &
                &  NoiseCl(1:2,l), Nldd(l), fsky
        end do
     endif
  endif
  if (num_cls == 3) then
     if (DoMock == .false.) then 
        do l=lmin, lmax     
           write (34,'(1I5,6E15.5)') l, Bigcov(l,1,1:2), Bigcov(l,2,2), &
                &  NoiseCl(1:2,l), fsky
        end do
     else
        do l = lmin, lmax     
           write (34,'(1I5,6E15.5)') l, Bigcov_mock(l,1,1:2), Bigcov_mock(l,2,2), &
                &  NoiseCl(1:2,l), fsky
        end do
     endif
  endif
  close(34)
  
  !
  ! Save memory
  !____________________________________
  deallocate(NoiseCl)
  deallocate(Nldd)
  if (DoMock) deallocate(Bigcov_mock)
  deallocate(Bigcov)
  
  setoutname = trim(datarep)//trim(setname)//".dataset"
  if (chatty == 1) print *, code//'> writing the dataset file: '//trim(setoutname)
  call CreateTxtFile(setoutname,37)
  write (37,'(A,A)')  "name = ",trim(setname)
  write (37,'(A,L1)') "all_l_exact = ",.true.
  write (37,'(A,L1)') "Do_Mock = ", DoMock 
  write (37,'(A,I5)') "mock_ncmbcls = ",num_cls
  if (num_cls>=3) write (37,'(A,L1)')  "has_pol = ",.true.
  write (37,'(A,A)')  "CMBCl_theo = ",trim(cmbcl_infile)
  if (num_cls == 5) write (37,'(A,A)')  "CMBLensCl_theo = ",trim(cmbcllens_infile)
  write (37,'(A,I5)') "all_l_lmax = ",lmax
  write (37,'(A,I5)') "all_l_lmin = ",lmin
  write (37,'(A,A)')  "all_l_file = ",trim(outname)
  close(37)
  
  
  
end program FuturCMB







